"""Node sets for the `d`-simplex implemented by ``recursivenodes``"""

import numpy as np

from recursivenodes.utils import npolys, multiindex_equal, multiindex_up_to, coord_map
from recursivenodes.quadrature import gaussjacobi, lobattogaussjacobi
from recursivenodes.polynomials import proriolkoornwinderdubinervandermonde

def equispaced(d, n, domain='biunit'):
    '''Equispaced nodes for polynomials up to degree `n` on the `d`-simplex.

    Args:
        d (int): The dimension of the simplex.
        n (int): The polynomial degree
        domain (str, optional) -- The domain of the simplex.  See ":ref:`domains`" for the choices and their formal definitions.

    Returns:
        ndarray: Equispaced nodes as a 2D array with one row for each of
        `\\binom{n+d}{d}` nodes, and `d` columns for coordinates (or `d+1`
        if ``domain='barycentric'``).

    Example:

        .. plot::
           :include-source: True

           >>> import matplotlib.pyplot as plt
           >>> from numpy import eye
           >>> from recursivenodes.nodes import equispaced
           >>> from recursivenodes.utils import coord_map
           >>> nodes = equispaced(2, 7, domain='equilateral')
           >>> corners = coord_map(eye(3), 'barycentric', 'equilateral')
           >>> plt.plot(corners[[0,1,2,0],0], corners[[0,1,2,0],1])
           [<matplotlib.lines.Line2D object at ...>]
           >>> plt.scatter(nodes[:,0], nodes[:,1])
           <matplotlib.collections.PathCollection object at ...>
           >>> plt.gca().set_aspect('equal')
           >>> plt.title('Equispaced Nodes')
           Text(0.5, 1.0, 'Equispaced Nodes')
           >>> plt.show()
    '''
    N = npolys(d, n)
    x = np.ndarray((N, d))
    for (k, i) in enumerate(multiindex_up_to(d, n)):
        x[k, :] = np.array(i) / n
    return coord_map(x, 'unit', domain)


def equispaced_interior(d, n, domain='biunit'):
    '''Equispaced nodes for polynomials up to degree `n` on the `d`-simplex,
    all of which are in the interior.

    Args:
        d (int): The dimension of the simplex.
        n (int): The polynomial degree
        domain (str, optional): The domain of the simplex.  See ":ref:`domains`" for the choices and their formal definitions.

    Returns:
        ndarray: Equispaced interior nodes as a 2D array with one row for each of
        `\\binom{n+d}{d}` nodes, and `d` columns for coordinates (or `d+1`
        if ``domain='barycentric'``).

    Example:

        .. plot::
           :include-source: True

           >>> import matplotlib.pyplot as plt
           >>> from numpy import eye
           >>> from recursivenodes.nodes import equispaced_interior
           >>> from recursivenodes.utils import coord_map
           >>> nodes = equispaced_interior(2, 7, domain='equilateral')
           >>> corners = coord_map(eye(3), 'barycentric', 'equilateral')
           >>> plt.plot(corners[[0,1,2,0],0], corners[[0,1,2,0],1])
           [<matplotlib.lines.Line2D object at ...>]
           >>> plt.scatter(nodes[:,0], nodes[:,1])
           <matplotlib.collections.PathCollection object at ...>
           >>> plt.gca().set_aspect('equal')
           >>> plt.title('Equispaced Interior Nodes')
           Text(0.5, 1.0, 'Equispaced Interior Nodes')
           >>> plt.show()
    '''
    N = npolys(d, n)
    x = np.ndarray((N, d))
    for (k, i) in enumerate(multiindex_up_to(d, n)):
        x[k, :] = (np.array(i) + 1/(d+1)) / (n+1)
    return coord_map(x, 'unit', domain)


def blyth_luo_pozrikidis(d, n, x=None, domain='biunit'):
    '''Create Blyth-Luo-Pozrikidis nodes from a 1D node set for polynomials up
    to degree `n` on the `d`-simplex.

    Notes:

        The Blyth-Luo-Pozrikidis rule places an interior node with multi-index
        `\\alpha` at the barycentric point `\\boldsymbol{b}(\\alpha)` such that

        .. math::
           :label: blp

           b_i(\\alpha) = x^n_{\\alpha_i} + \\frac{1}{d}(1 - \\sum_{j\\neq i} x^n_{\\alpha_j}).

        Points on the boundary look like `(d-1)`-simplex Blyth-Luo-Pozrikidis nodes.

    Args:
        d (int): The dimension of the simplex.
        n (int): The polynomial degree
        x (ndarray, optional): 1D node set on `[0, 1]` with `n+1` points.  Lobatto-Gauss-Legendre nodes are used if  ``x=None``.
        domain (str, optional): The domain of the simplex.  See ":ref:`domains`" for the choices and their formal definitions.

    Returns:
        ndarray: Blyth-Luo-Pozrikidis nodes as a 2D array with one row for each of
        `\\binom{n+d}{d}` nodes, and `d` columns for coordinates (or `d+1`
        if ``domain='barycentric'``).

    Example:

        This plot shows the Blyth-Luo-Pozrikidis nodes with lines connecting the
        Lobatto-Gauss-Legendre nodes on the edges.  The definition in :eq:`blp` is
        designed to place the nodes in the centroids of the triangles created by the
        intersecting lines.

        .. plot::
           :include-source: True

           >>> import matplotlib.pyplot as plt
           >>> from numpy import eye
           >>> from recursivenodes.nodes import blyth_luo_pozrikidis
           >>> from recursivenodes.utils import coord_map, multiindex_equal
           >>> nodes = blyth_luo_pozrikidis(2, 7, domain='equilateral')
           >>> corners = coord_map(eye(3), 'barycentric', 'equilateral')
           >>> # plot grid lines
           >>> for (i, al) in enumerate(multiindex_equal(3, 7)):
           ...     if min(al) > 0 or max(al) == sum(al): continue
           ...     for (j, be) in enumerate(multiindex_equal(3, 7)):
           ...         if j <= i or min(be) > 0 or max(be) != max(al) or sum(be) != sum(al): continue
           ...         for d in range(3):
           ...             if al[d] == be[d] and al[d] != 0:
           ...                 _ = plt.plot(nodes[[i,j],0], nodes[[i,j],1], linestyle='--', c='grey')
           >>> plt.plot(corners[[0,1,2,0],0], corners[[0,1,2,0],1])
           [<matplotlib.lines.Line2D object at ...>]
           >>> plt.scatter(nodes[:,0], nodes[:,1])
           <matplotlib.collections.PathCollection object at ...>
           >>> plt.gca().set_aspect('equal')
           >>> plt.title('Blyth-Luo-Pozrikidis Nodes')
           Text(0.5, 1.0, 'Blyth-Luo-Pozrikidis Nodes')
           >>> plt.show()

    References:

        :cite:`BlPo06,LuPo06`
    '''
    if x is None:
        x = coord_map(lobattogaussjacobi(n+1)[0], 'biunit', domain)
    n = len(x) - 1
    y = coord_map(x, domain, 'unit')
    N = npolys(d, n)
    xd = np.zeros((N, d))
    k = 0
    for (k, i) in enumerate(multiindex_equal(d+1, n)):
        numzero = sum([int(y[l] == 0.) for l in i])
        weight = 0.
        for j in range(d+1):
            ibutj = i[0:j] + i[(j+1):]
            yibutj = [y[l] for l in ibutj]
            thisnumzero = sum([int(z == 0) for z in yibutj])
            if thisnumzero < numzero: continue
            yj = 1. - sum(yibutj)
            weight += 1
            for l in range(j):
                if (l < d):
                    xd[k, l] += y[i[l]]
            if (j < d):
                xd[k, j] += yj
            for l in range(j+1, d+1):
                if (l < d):
                    xd[k, l] += y[i[l]]
        xd[k,:] /= weight
    return coord_map(xd, 'unit', domain)


def _warburton_b(d, xt, i):
    b = np.ones((xt.shape[0], 1))
    tol = 1.e-8
    bdry = np.zeros(b.shape)
    for j in range(d+1):
        if j == i:
            continue
        nonzero = np.abs(2. * xt[:, j] + xt[:, i]) > tol
        b[nonzero, 0] *= (2. * xt[nonzero, j]) / (2. * xt[nonzero, j] + xt[nonzero, i])
        bdry[~nonzero, 0] += 1.
    isbdry = bdry[:, 0] > 0.
    if d == 2:
        ''' The implementation of warp and blend, evident in the Lebesgue constants
        in the paper and in modepy (https://github.com/inducer/modepy/blob/master/modepy/nodes.py)
        uses the blending scaling (2 * l2) * (2 * l3) / (1. - (l2 - l3)**2) instead of
        (2 * l2) * (2 * l3) / ((2 * l2 + l1)*(2 * l3 + l1)).

        If l1 + l2 + l3 = 1, this is equivalent, because

        (1. - (l2 - l3)**2) = (1. + (l2 - l3))*(1. - (l2 - l3))
                            = (l1+l2+l3 +l2-l3)*(l1+l2+l3 -l2+l3)=(2*l2+l1)*(2.*l3+l1)

        BUT, when we are blending for higher dimensions, we may be using a subset of
        barycentric coordinates, i.e. l1 + l2 + l3 + l4 = 1, l4 > 0, and we are computing
        a face warp for the l4 face, so we use only the (l1,l2,l3) coordinate.  In this case
        l1+l2+l3 != 1, so the two blending scalings are not equivalent.  I'm choosing
        to follow the implementation so that I can reproduce the computed Lebesgue constants.'''
        i2 = (i + 1) % 3
        i3 = (i + 2) % 3
        b[~isbdry, 0] = 4. * xt[~isbdry, i2] * xt[~isbdry, i3] \
                        / (1. - (xt[~isbdry, i2]-xt[~isbdry, i3])**2)
    b[isbdry, 0] = 1. / (bdry[isbdry, 0] + 1.)
    return b


def _warburton_g(d, n, xt, alpha=0., x=None):
    g = np.zeros(xt.shape)
    if (d == 1):
        if x is None:
            x, _ = lobattogaussjacobi(n+1)
        xe = np.linspace(-1., 1., n+1, endpoint=True)
        diff = (x - xe) / 2
        V = proriolkoornwinderdubinervandermonde(1, n, xe.reshape((n+1, 1)))
        P = np.linalg.solve(V, diff)
        T = proriolkoornwinderdubinervandermonde(1, n, xt[:, [0]] - xt[:, [1]])
        g[:, 0] = T.dot(P)
        g[:, 1] = -g[:, 0]
    else:
        for i in range(d+1):
            rem = [j for j in range(d+1) if j != i]
            gi = _warburton_g(d-1, n, xt[:, rem], alpha, x)
            bi = _warburton_b(d, xt, i)
            g[:, rem] += (1. + (alpha * xt[:, [i]])**2) * gi * bi
    return g


#: A dictionary of optimal values of the blending parameter `\alpha` computed
#: in :cite:`Warb06`.  Keyed by ``(d, n)`` tuples.
#: 
#: Example:
#:     >>> warburton_alpha[(2,7)]
#:     1.0999
warburton_alpha = {
    (2, 3):  1.4152, (3, 3):  0.0000,
    (2, 4):  0.1001, (3, 4):  0.1002,
    (2, 5):  0.2751, (3, 5):  1.1332,
    (2, 6):  0.9808, (3, 6):  1.5608,
    (2, 7):  1.0999, (3, 7):  1.3413,
    (2, 8):  1.2832, (3, 8):  1.2577,
    (2, 9):  1.3648, (3, 9):  1.1603,
    (2, 10): 1.4773, (3, 10): 1.0153,
    (2, 11): 1.4959, (3, 11): 0.6080,
    (2, 12): 1.5743, (3, 12): 0.4523,
    (2, 13): 1.5770, (3, 13): 0.8856,
    (2, 14): 1.6223, (3, 14): 0.8717,
    (2, 15): 1.6258, (3, 15): 0.9655,
}


def warburton(d, n, x=None, alpha=None, domain='biunit'):
    '''Warburton *warp & blend* nodes from a 1D node set for polynomials up
    to degree `n` on the `d`-simplex.

    Notes:

        The Warburton *warp & blend* nodes define a node's coordinates as a
        displacement from equispaced coordinates by blending together
        distortion maps of the `d`-simplex that warp the edge nodes
        to match the 1D node set.  One optimization parameter `\\alpha`
        controls the blending in the interior of the simplex.

    Args:
        d (int): The dimension of the simplex.
        n (int): The polynomial degree
        x (ndarray, optional): 1D node set on `[0, 1]` with `n+1` points.  Lobatto-Gauss-Legendre nodes are used if ``x=None``.
        alpha (float, optional): The blending parameter. If ``alpha=None``, a precomputed optimal parameter is used if known.
        domain (str, optional): The domain of the simplex.  See ":ref:`domains`" for the choices and their formal definitions.

    Returns:
        ndarray: Warburton *warp & blend* nodes as a 2D array with one row for each of
        `\\binom{n+d}{d}` nodes, and `d` columns for coordinates (or `d+1`
        if ``domain='barycentric'``).

    Example:

        .. plot::
           :include-source: True

           >>> import matplotlib.pyplot as plt
           >>> from numpy import eye
           >>> from recursivenodes.nodes import warburton
           >>> from recursivenodes.utils import coord_map
           >>> nodes = warburton(2, 7, domain='equilateral')
           >>> corners = coord_map(eye(3), 'barycentric', 'equilateral')
           >>> plt.plot(corners[[0,1,2,0],0], corners[[0,1,2,0],1])
           [<matplotlib.lines.Line2D object at ...>]
           >>> plt.scatter(nodes[:,0], nodes[:,1])
           <matplotlib.collections.PathCollection object at ...>
           >>> plt.gca().set_aspect('equal')
           >>> plt.title('Warburton Warp & Blend Nodes')
           Text(0.5, 1.0, 'Warburton Warp & Blend Nodes')
           >>> plt.show()

    References:

        :cite:`Warb06`
    '''
    if not (x is None):
        x = coord_map(x, domain, 'biunit')
    if alpha is None:
        try:
            alpha = warburton_alpha[(d, n)]
        except KeyError:
            alpha = 0.
    xt = equispaced(d, n, domain='barycentric')
    g = _warburton_g(d, n, xt, alpha, x)
    xt += g
    return coord_map(xt, 'barycentric', domain)


class NodeFamily:
    ''' Family of nodes on the unit interval.  This class essentially is a lazy-evaluate-and-cache dictionary: the user passes a routine to evaluate entries for unknown keys '''

    def __init__(self, f):
        self._f = f
        self._cache = {}

    def __getitem__(self, key):
        try:
            return self._cache[key]
        except KeyError:
            value = self._f(key)
            self._cache[key] = value
            return value


# For each family, family[n] should be a a set of n+1 points in [0,1] in increasing order that is symmetric about 1/2,
# family[n] should be None if the family does not have a representative for that index
#
# We predefine:
#
#   - shifted Lobatto-Gauss-Legendre  (lgl_family)
#   - shifted Lobatto-Gauss-Chebyshev (lgc_family)
#   - shifted Gauss-Legendre          (gl_family)
#   - shifted Gauss-Chebyshev         (gc_family)
#   - equispaced, including endpoints (equi_family)
#   - equispaced, interior            (equi_interior_family)
lgl_family = NodeFamily(lambda n: coord_map(lobattogaussjacobi(n+1)[0], 'biunit', 'unit') if n > 0 else None)
lgc_family = NodeFamily(lambda n: coord_map(lobattogaussjacobi(n+1, -0.5, -0.5)[0], 'biunit', 'unit') if n > 0 else None)
gl_family = NodeFamily(lambda n: coord_map(gaussjacobi(n+1)[0], 'biunit', 'unit') if n >= 0 else None)
gc_family = NodeFamily(lambda n: coord_map(gaussjacobi(n+1, -0.5, -0.5)[0], 'biunit', 'unit') if n >= 0 else None)
equi_family = NodeFamily(lambda n: np.linspace(0., 1., n+1) if n > 0 else None)
equi_interior_family = NodeFamily(lambda n: np.linspace(
    0.5/(n+1), 1.-0.5/(n+1), n+1) if n >= 0 else None)


def _recursive(d, n, alpha, family):
    '''The barycentric d-simplex coordinates for a
    multiindex alpha with length n, based on a 1D node family.'''
    xn = family[n]
    b = np.zeros((d,))
    if xn is None:
        return b
    if d == 2:
        b[:] = xn[[alpha[0],alpha[1]]] 
        return b
    weight = 0.
    for i in range(d):
        alpha_noti = alpha[:i] + alpha[i+1:]
        n_noti = n - alpha[i]
        w = xn[n_noti]
        br = _recursive(d-1, n_noti, alpha_noti, family)
        b[:i] += w * br[:i]
        b[i+1:] += w * br[i:]
        weight += w
    b /= weight
    return b


def _decode_family(family):
    if family is None:
        family = lgl_family
    elif isinstance(family, str):
        if family == 'lgl':
            family = lgl_family
        elif family == 'lgc':
            family = lgc_family
        elif family == 'gl':
            family = gl_family
        elif family == 'gc':
            family = gc_family
        elif family == 'equi':
            family = equi_family
        elif family == 'equi_interior':
            family = equi_interior_family
    elif isinstance(family, tuple) and len(family) == 2 and family[0] == 'lgg':
        a = family[1]
        family = NodeFamily(lambda n: coord_map(lobattogaussjacobi(n+1, a-0.5, a-0.5)[0], 'biunit', 'unit') if n > 0 else None)
    elif isinstance(family, tuple) and len(family) == 2 and family[0] == 'gg':
        a = family[1]
        family = NodeFamily(lambda n: coord_map(gaussjacobi(n+1, a-0.5, a-0.5)[0], 'biunit', 'unit') if n >= 0 else None)
    return family


def recursive(d, n, family='lgl', domain='barycentric'):
    '''Recursively defined nodes for `\\mathcal{P}_n(\\Delta^d)`, the polynomials
    with degree at most `n` on the `d`-simplex, based on a 1D node family.

    Notes:

        Some definitions:

        - A 1D *node family* is a collection `\\{\\boldsymbol{x}^k\\}` of 1D node sets
          for every degree `k`.

        - A 1D *node set* `\\boldsymbol{x}^k=(x^k_0, \\dots, x^k_k)` is a sorted list
          of `k+1` points in `[0,1]` that is symmetric about `1/2`.

        - The `barycentric triangle`_ is a canonical domain for the `d`-simplex in
          `\\mathbb{R}^{d+1}`: positive coordinates `\\boldsymbol{b}
          = (b_0, b_1, \\dots b_d)` such that `\\sum_i b_i = 1`.

        - For a multi-index_ `\\alpha`, let `\\#\\alpha` be its length,
          `|\\alpha|` its sum, and `\\alpha_{\\backslash i}` the multi-index
          created by removing `\\alpha_i`: nodes that define a basis of
          `\\mathcal{P}_n(\\Delta^d)` can be indexed by `\\alpha` such that
          `\\#\\alpha = d+1` and `|\\alpha| = n`.

        - `I^d_i` is the map from `\\mathbb{R}^{d-1}` to `\\mathbb{R}^d`
          that inserts a zero for the `i`-th coordinate.

        The recursive definition of barycentric node coordinates,
        `\\boldsymbol{b}_{\\boldsymbol{x}}(\\alpha) \\in \\mathbb{R}^{\\#\\alpha}`
        has the base case

        .. math::
            :label: basecase

            \\boldsymbol{b}_{\\boldsymbol{x}}(\\alpha) = (x^{|\\alpha|}_{\\alpha_0}, x^{|\\alpha|}_{\\alpha_1}), \\quad \\#\\alpha = 2,

        and the recursion

        .. math::
            :label: recursion

            \\boldsymbol{b}_{\\boldsymbol{x}}(\\alpha) =
            \\frac{\\sum_i x^{|\\alpha|}_{|\\alpha_{\\backslash i}|} I^{\\#\\alpha}_i \\boldsymbol{b}_{\\boldsymbol{x}}(\\alpha_{\\backslash i})}
            {\\sum_i x^{|\\alpha|}_{|\\alpha_{\\backslash i}|}}, \\quad \\#\\alpha > 2.

        The full set of nodes is

        .. math:: \\boldsymbol{b}_{\\boldsymbol{x}}^{d,n} = \\{\\boldsymbol{b}_{\\boldsymbol{x}}(\\alpha): \\#\\alpha = d+1, |\\alpha| = n\\}.
            :label: fullset

    Args:
        d (int): The dimension of the simplex
        n (int): The maximum degree of the polynomials
        family (optional): The 1D node family used to define the coordinates in the barycentric `d`-simplex.
            The default ``family='lgl'`` corresponds to the shifted Lobatto-Gauss-Legendre_ nodes.
            See ":ref:`families`" for using other node families.
        domain (str, optional): The domain for the `d`-simplex where the returned coordinates will be defined.
            See ":ref:`domains`" for the choices and their formal definitions.

    Returns:
        ndarray: The nodes `\\boldsymbol{b}_{\\boldsymbol{x}}^{d,n}` defined in
        :eq:`fullset`, as a 2D array with `\\binom{n+d}{d}` rows.  If
        ``domain='barycentric'``, it has `d+1` columns, otherwise it has `d`
        columns.

    Examples:

        Nodes for `\\mathcal{P}^4(\\Delta^2)` in barycentric coordinates:

        >>> recursive_nodes(2, 4)
        array([[0.        , 0.        , 1.        ],
               [0.        , 0.17267316, 0.82732684],
               [0.        , 0.5       , 0.5       ],
               [0.        , 0.82732684, 0.17267316],
               [0.        , 1.        , 0.        ],
               [0.17267316, 0.        , 0.82732684],
               [0.2221552 , 0.2221552 , 0.5556896 ],
               [0.2221552 , 0.5556896 , 0.2221552 ],
               [0.17267316, 0.82732684, 0.        ],
               [0.5       , 0.        , 0.5       ],
               [0.5556896 , 0.2221552 , 0.2221552 ],
               [0.5       , 0.5       , 0.        ],
               [0.82732684, 0.        , 0.17267316],
               [0.82732684, 0.17267316, 0.        ],
               [1.        , 0.        , 0.        ]])

        The same nodes on the unit triangle:

        >>> recursive_nodes(2, 4, domain='unit')
        array([[0.        , 0.        ],
               [0.        , 0.17267316],
               [0.        , 0.5       ],
               [0.        , 0.82732684],
               [0.        , 1.        ],
               [0.17267316, 0.        ],
               [0.2221552 , 0.2221552 ],
               [0.2221552 , 0.5556896 ],
               [0.17267316, 0.82732684],
               [0.5       , 0.        ],
               [0.5556896 , 0.2221552 ],
               [0.5       , 0.5       ],
               [0.82732684, 0.        ],
               [0.82732684, 0.17267316],
               [1.        , 0.        ]])

        If we construct the node set not from the Lobatto-Gauss-Legendre 1D node family,
        but from the equispaced 1D node family, we get equispaced 2D nodes:

        >>> recursive_nodes(2, 4, family='equi', domain='unit')
        array([[0.  , 0.  ],
               [0.  , 0.25],
               [0.  , 0.5 ],
               [0.  , 0.75],
               [0.  , 1.  ],
               [0.25, 0.  ],
               [0.25, 0.25],
               [0.25, 0.5 ],
               [0.25, 0.75],
               [0.5 , 0.  ],
               [0.5 , 0.25],
               [0.5 , 0.5 ],
               [0.75, 0.  ],
               [0.75, 0.25],
               [1.  , 0.  ]])

        This is what they look like mapped to the equilateral triangle:

        .. plot::
           :include-source: True

           >>> import matplotlib.pyplot as plt
           >>> from recursivenodes import recursive_nodes
           >>> nodes_equi = recursive_nodes(2, 4, family='equi', domain='equilateral')
           >>> nodes_lgl = recursive_nodes(2, 4, domain='equilateral')
           >>> plt.scatter(nodes_lgl[:,0], nodes_lgl[:,1], marker='o', label='recursive LGL')
           <matplotlib.collections.PathCollection object at ...>
           >>> plt.scatter(nodes_equi[:,0], nodes_equi[:,1], marker='^', label='equispaced')
           <matplotlib.collections.PathCollection object at ...>
           >>> plt.gca().set_aspect('equal')
           >>> plt.legend()
           <matplotlib.legend.Legend object at ...>
           >>> plt.show()


    .. _multi-index: https://en.wikipedia.org/wiki/Multi-index_notation
    .. _barycentric triangle: https://en.wikipedia.org/wiki/Barycentric_coordinate_system
    .. _Lobatto-Gauss-Legendre: https://en.wikipedia.org/wiki/Gaussian_quadrature#Gauss%E2%80%93Lobatto_rules

    '''
    family = _decode_family(family)
    N = npolys(d, n)
    x = np.zeros((N, d+1))
    for (k, i) in enumerate(multiindex_equal(d+1, n)):
        x[k, :] = _recursive(d+1, n, i, family)
    return coord_map(x, 'barycentric', domain)


def expand_to_boundary(d, n, nodes, ex_nodes, domain='biunit'):
    ''' given degree n nodes for the d-simplex that are logically laid out like
    the equispaced nodes and are only in the interior, construct degree (n + d
    + 1) nodes for the d-simplex which are equal to nodes in the interior and
    ex_nodes on the boundary.  No attempt is made to make the new boundary
    nodes good for interpolation: they are just there to aid in computing
    lebesguemax() in lebesgue.py '''
    nd = n + d + 1
    tuple_to_index = {}
    for (k, i) in enumerate(multiindex_equal(d+1, n)):
        tuple_to_index[i] = k
    ex_nodes = ex_nodes.copy()
    for (k, i) in enumerate(multiindex_equal(d+1, nd)):
        if min(i) > 0:
            ihat = tuple([a - 1 for a in i])
            j = tuple_to_index[ihat]
            ex_nodes[k, :] = nodes[j, :]
    return ex_nodes


def add_nodes_to_parser(parser):
    parser.add_argument('-n', '--degree', type=int, nargs='?', default=7,
                        help='degree of polynomial space for which node set is equisolvent')
    parser.add_argument('-d', '--dimension', type=int, nargs='?', default=3,
                        help='spatial dimensions to node set simplex')
    parser.add_argument('--nodes', type=str,
                        choices=['recursive', 'warburton', 'blyth_luo_pozrikidis', 'equispaced', 'equispaced_interior'],
                        nargs='?', default='recursive',
                        help='node placement algorithm')
    parser.add_argument('-w', '--w-alpha', type=float, nargs='?',
                        help='blending parameter for Warburton warp & blend method')
    parser.add_argument('--domain', type=str,
                        choices=['biunit', 'unit', 'barycentric', 'equilateral'],
                        default='biunit', nargs='?',
                        help='output simlex domain')
    parser.add_argument('-f', '--family', type=str, choices=['lgl', 'lgc', 'gl', 'gc', 'equi', 'equi_interior', 'lgg', 'gg'],
                        default='lgl', nargs='?',
                        help='base 1D node family for Blyth-Pozrikidis, Warburton, and recursive nodes')
    parser.add_argument('-g', '--g-alpha', type=float, nargs='?', default=0.,
                        help='Gegenbauer alpha exponent (only used with --family lgg or --family gg)')


def nodes_from_args(args, expanded=False):
    d = args.dimension
    n = args.degree
    domain = args.domain
    fam_arg = args.family
    need_expanded = False
    if expanded:
        need_expanded = True
        if args.nodes == 'equispaced' or (args.nodes != 'equispaced_interior' and (args.family == 'equi' or fam_arg[0] == 'l')):
            need_expanded = False
    if need_expanded:
        if args.nodes == 'equispaced_interior' or args.family == 'equi_interior':
            ex_fam_arg = 'equi'
        else:
            ex_fam_arg = 'l' + fam_arg
    if (fam_arg == 'lgg' or fam_arg == 'gg'):
        fam_arg = (fam_arg, args.g_alpha)
    family = _decode_family(fam_arg)
    x = coord_map(family[n], 'unit', domain)
    if need_expanded:
        if (ex_fam_arg == 'lgg'):
            ex_fam_arg = (ex_fam_arg, args.g_alpha)
        ex_family = _decode_family(ex_fam_arg)
        ex_x = coord_map(family[n+d+1], 'unit', domain)
    if args.nodes == 'equispaced':
        nodes = equispaced(d, n, domain=domain)
    elif args.nodes == 'equispaced_interior':
        nodes = equispaced_interior(d, n, domain=domain)
        if need_expanded:
            ex_nodes = equispaced(d, n+d+1, domain=domain)
    elif args.nodes == 'blyth_luo_pozrikidis':
        nodes = blyth_luo_pozrikidis(d, n, x, domain=domain)
        if need_expanded:
            ex_nodes = blyth_luo_pozrikidis(d, n+d+1, ex_x, domain=domain)
    elif args.nodes == 'warburton':
        alpha = args.w_alpha
        nodes = warburton(d, n, x, alpha=alpha, domain=domain)
        if need_expanded:
            ex_nodes = warburton(d, n+d+1, ex_x, alpha=alpha, domain=domain)
    elif args.nodes == 'recursive':
        nodes = recursive(d, n, family, domain=domain)
        if need_expanded:
            ex_nodes = recursive(d, n+d+1, ex_family, domain=domain)
    if expanded:
        if need_expanded:
            ex_nodes = expand_to_boundary(d, n, nodes, ex_nodes, domain=domain)
            return (d, n, domain, family, nodes, ex_nodes)
        else:
            return (d, n, domain, family, nodes, nodes)
    return (d, n, domain, family, nodes)


if __name__ == '__main__':
    from argparse import ArgumentParser
    parser = ArgumentParser(description='Print out a node set implemented in nodes.py')
    add_nodes_to_parser(parser)
    args = parser.parse_args()
    d, n, domain, family, nodes = nodes_from_args(args)
    print(nodes)



"""Utilities used by ``recursivenodes``.

.. testsetup:: *

    from recursivenodes.utils import *
"""

import numpy as np
from scipy.special import comb


def multiindex_equal(d, k):
    """A generator for :math:`d`-tuple multi-indices whose sum is :math:`k`.

    Args:
        d (int): The length of the tuples
        k (int): The sum of the entries in the tuples

    Yields:
        tuple: tuples of length `d` whose entries sum to `k`, in lexicographic
        order.

    Example:
        >>> for i in multiindex_equal(3, 2): print(i)
        (0, 0, 2)
        (0, 1, 1)
        (0, 2, 0)
        (1, 0, 1)
        (1, 1, 0)
        (2, 0, 0)
    """
    if d <= 0:
        return
    if k < 0:
        return
    for i in range(k):
        for a in multiindex_equal(d-1, k-i):
            yield (i,) + a
    yield (k,) + (0,)*(d-1)


def multiindex_up_to(d, k):
    """A generator for :math:`d`-tuple multi-indices whose sum is at most
    :math:`k`.

    Args:
        d (int): The length of the tuples
        k (int): The maximum sum of the entries in the tuples

    Yields:
        tuple: tuples of length `d` whose entries sum to `k`, in lexicographic
        order.
    
    Example:
        >>> for i in multiindex_up_to(3, 2): print(i)
        (0, 0, 0)
        (0, 0, 1)
        (0, 0, 2)
        (0, 1, 0)
        (0, 1, 1)
        (0, 2, 0)
        (1, 0, 0)
        (1, 0, 1)
        (1, 1, 0)
        (2, 0, 0)
    """
    for a in multiindex_equal(d+1, k):
        yield a[0:d]


def npolys(d, k):
    """The number of polynomials up to a degree :math:`k` in :math:`d` dimensions.

    Args:
        d (int): The dimension of the space.
        k (int): The polynomial degree.

    Returns:
        int: :math:`\\binom{k+d}{d}`.

    Example:
        >>> npolys(3, 2)
        10
    """
    return int(comb(k+d, d))


def _equilateral_to_unit(d, x):
    if d > 1:
        # scale the vertical dimension
        x[..., d-1] /= ((d+1.)/(2*d))**0.5
        # make the projection onto the lesser dimensions right-angled
        _equilateral_to_unit(d-1, x[..., 0:d-1])
        x[..., 0:d-1] -= x[..., d-1, np.newaxis] / d  # move the top vertex over first vertex


def _to_unit(x, domain):
    if domain == 'unit':
        return x.copy()
    if domain == 'biunit':
        return (x + 1) / 2
    if domain == 'barycentric':
        return x[..., 0:-1].copy()
    if domain == 'equilateral':
        d = x.shape[-1]
        z = x.copy()
        # scale edge length to 1
        z /= 2
        _equilateral_to_unit(d, z)
        z += 1/(d+1)  # shift the first vertex to zero
        return z
    else:
        raise NotImplementedError


def _unit_to_equilateral(d, x):
    if d > 1:
        x[..., 0:d-1] += x[..., d-1, np.newaxis] / d  # move the top vertex over the centroid
        # make the projection onto the lesser dimensions equilateral
        _unit_to_equilateral(d-1, x[..., 0:d-1])
        # scale the vertical dimension
        x[..., d-1] *= ((d+1.)/(2*d))**0.5


def _from_unit(x, domain):
    if domain == 'unit':
        return x.copy()
    if domain == 'biunit':
        return x*2 - 1
    if domain == 'barycentric':
        z = 1 - x.sum(axis=-1)
        return np.concatenate((x, z[..., np.newaxis]), axis=-1)
    if domain == 'equilateral':
        d = x.shape[-1]
        z = x.copy()
        z -= 1/(d+1)  # shift centroid to zero
        _unit_to_equilateral(d, z)
        # scale edge length to 2
        z *= 2
        return z
    else:
        raise NotImplementedError

def coord_map(x, origin, image):
    '''Map coordinates from one reference `d`-simplex to another.

    Args:
        x (ndarray): 2D array of coordinates of shape ``(num_points, coordinate_dimension)``
        origin (str): Current domain of `x`, one of ``['unit', 'biunit', 'equilateral', 'barycentric']``.
        image (str): Desired domain of mapped coordinates.

    Returns:
        ndarray: 2D array of coordinates in the new domain, shape ``(num_points, new_coordinated_dimension)``.

    See Also:
        ":ref:`domains`" for formal definitions of the domains.
    '''
    return _from_unit(_to_unit(x, domain=origin), domain=image)



'''Quadrature nodes used in ``recursivenodes``.'''

import numpy as np
from scipy.special import gamma
from scipy.linalg import eigh_tridiagonal
from itertools import product
from recursivenodes.polynomials import jacobi, jacobider, jacobinorm2
from recursivenodes.utils import multiindex_equal, multiindex_up_to, npolys


def _jacobi_matrix(n, a=0., b=0.):
    ''' return as numpy arrays, the diagonals for the Jacobi recurrence relationship '''
    alpha = np.ndarray((n,))
    beta = np.ndarray((n,))
    for i in range(1, n+1):
        if i > 1:
            A = (2*i+a+b)*(2*i+a+b-1) / (2*i*(i+a+b))
            B = (a*a-b*b)*(2*i+a+b-1) / (2*i*(i+a+b)*(2*i+a+b-2))
        else:
            A = (a+b+2)/2
            B = (a-b)/2
        Ap = (2*(i+1)+a+b)*(2*(i+1)+a+b-1) / (2*(i+1)*(i+1+a+b))
        C = 2*((i+1)+a-1)*((i+1)+b-1)*(2*(i+1)+a+b) / (2*(i+1)*((i+1)+a+b)*(2*(i+1)+a+b-2))
        alpha[i-1] = -B / A
        beta[i-1] = (C / (A * Ap))
    return (alpha, beta)


def _monicjacobi(n, x, alpha, beta):
    ''' evaluate the monic Jacobi polynomial:
        - for the interval [-1,1]
        - with degree n
        - with the diagonals generated by _jacobi_matrix(n,a,b) for the Jacobi weight (1+x)^a (1-x)^b
        - at the points x (a numpy array)

        returns a numpy array with the same shape as x
    '''
    pm2 = np.ones(np.shape(x))
    pm1 = (x - alpha[0]) * pm2
    for i in range(1, n):
        pm2 *= -beta[i-1]
        pm2 += (x - alpha[i]) * pm1
        p = pm2
        pm2 = pm1
        pm1 = p
    return pm1


def gaussjacobi(n, a=0., b=0.):
    '''The `n`-point Gauss-Jacobi quadrature points and weights for weight function `(1+x)^a (1-x)^b` for the interval [-1,1],
    which exactly integrates polynomials with degree up to `2n-1`.

    Args:
        n (int): The number of points.
        a (float, optional): the left exponent of the weight function.
        b (float, optional): the right exponent of the weight function.

    Returns:
        (ndarray, ndarray): Each of shape (`n`,), the points (in ascending order) and weights of the quadrature rule.

    Reference:
        :cite:`GoWe69`
    '''
    alpha, beta = _jacobi_matrix(n, a, b)
    x, V = eigh_tridiagonal(alpha, (beta[:-1])**0.5)
    if (a == b):
        x = (x - x[::-1]) / 2.
        if (n % 2):
            x[n // 2] = 0.
    mu0 = 2**(a+b+1)*gamma(a+1)*gamma(b+1) / gamma(a+b+2)
    w = V[0, :]**2. * mu0
    if (a == b):
        w = (w + w[::-1]) / 2.
    return (x, w)


def lobattogaussjacobi(n, a=0., b=0.):
    '''The `n`-point Lobatto-Gauss-Jacobi quadrature points and weights for weight function `(1+x)^a (1-x)^b` for the interval [-1,1],
    which includes the endpoints `-1` and `1` and exactly integrates polynomials with degree up to `2n-3`.

    Args:
        n (int): The number of points.
        a (float, optional): the left exponent of the weight function.
        b (float, optional): the right exponent of the weight function.

    Returns:
        (ndarray, ndarray): Each of shape (`n`,), the points (in ascending order) and weights of the quadrature rule.

    Reference:
        :cite:`Golu73`
    '''
    mu0 = 2**(a+b+1)*gamma(a+1)*gamma(b+1) / gamma(a+b+2)
    if (n == 2):
        return (np.array([-1., 1.]), np.array([mu0/2., mu0/2.]))
    alpha, beta = _jacobi_matrix(n, a, b)
    ends = np.array([-1., 1.])
    p_nm2 = jacobi(n-2, ends, a, b)
    p_nm1 = jacobi(n-1, ends, a, b)
    p_nm2 = _monicjacobi(n-2, ends, alpha, beta)
    p_nm1 = _monicjacobi(n-1, ends, alpha, beta)
    rhs = np.array([-p_nm1[0], p_nm1[1]])
    Ainv = np.array([[p_nm2[1], -p_nm2[0]], [-p_nm1[1], p_nm1[0]]])
    Ainv /= (p_nm1[0] * p_nm2[1] - p_nm2[0] * p_nm1[1])
    sol = Ainv.dot(rhs)
    alpha[-1] = sol[0]
    beta[-2] = sol[1]
    x, V = eigh_tridiagonal(alpha, (beta[:-1])**0.5)
    if (a == b):
        x = (x - x[::-1]) / 2.
        if (n % 2):
            x[n // 2] = 0.
    w = V[0, :]**2. * mu0
    if (a == b):
        w = (w + w[::-1]) / 2.
    x[0] = -1.
    x[-1] = 1.
    return (x, w)


def simplexgausslegendre(d, n):
    '''An `n^d`-point Gaussian quadrature rule for the biunit simplex that exactly integrates polynomials with degree up to `2n-1`.

    Args:
        d (int): The spatial dimension.
        n (int): The points per coordinate direction

    Returns:
        (ndarray, ndarray): The points (shape (`n, \\dots, n, d`)) and weights (shape (`n, \\dots, n,`)) of the quadrature rule.
    '''
    if (d == 1):
        x, w = gaussjacobi(n, 0., 0.)
        return x.reshape((n, 1)), w
    xt, wt = simplexgausslegendre(d-1, n)
    xz, wz = gaussjacobi(n, d-1, 0.)
    x = np.ndarray((n,)*d+(d,))
    w = np.ndarray((n,)*d)
    for i in product(range(n), repeat=d-1):
        for j in range(n):
            x[i+(j,)][0:d-1] = (xt[i][0:d-1] + 1.) * ((1. - xz[j])/2.) - 1.
            x[i+(j,)][d-1] = xz[j]
            w[i+(j,)] = wt[i] * wz[j] / 2**(d-1)
    return x, w





'''Polynomials used in ``recursivenodes``.'''

import numpy as np
from scipy.special import gamma, eval_jacobi
from recursivenodes.utils import multiindex_up_to, npolys


def jacobi(n, x, a=0., b=0., out=None):
    '''Evaluation of a Jacobi polynomial.

    Args:
        n (int): The degree of the polynomial.
        x (ndarray): Points at which to evaluate the polynomial.
        a (float, optional): Left exponent of weight function.
        b (float, optional): Right exponent of weight function.
        out (ndarray, optional): Output placed here if given, same shape as ``x``.

    Returns:
        ndarray: same shape as ``x``, values of the `n`-th Jacobi polynomials
        with respect to the weight function `(1+x)^a(1-x)^b` at the points in
        ``x``.
    '''
    return eval_jacobi(n, a, b, x, out)


def jacobider(n, x, a=0., b=0., k=1, out=None):
    '''Evaluation of a the `k`-th derivative of a Jacobi polynomial.

    Args:
        n (int): The degree of the polynomial.
        x (ndarray): Points at which to evaluate the polynomial.
        a (float, optional): Left exponent of weight function.
        b (float, optional): Right exponent of weight function.
        out (ndarray, optional): Output placed here if given, same shape as ``x``.

    Returns:
        ndarray: same shape as ``x``, values of the `k`-th derivative of the
        `n`-th Jacobi polynomials with respect to the weight function
        `(1+x)^a(1-x)^b` at the points in ``x``.
    '''
    if k > n:
        if out:
            out[:] = 0.
            return out
        else:
            return np.zeros(np.shape(x))
    return jacobi(n-k, x, a+k, b+k) * gamma(a+b+n+1+k) / gamma(a+b+n+1) / 2**k


def jacobinorm2(n, a=0., b=0.):
    '''The square of the weighted `L_2` norm of a Jacobi polynomial.

    Args:
        n (int): The degree of the polynomial.
        x (ndarray): Points at which to evaluate the polynomial.
        a (float, optional): Left exponent of weight function.
        b (float, optional): Right exponent of weight function.

    Returns:
        float: `\\int_{-1}^1 (1+x)^a (1-x)^b P^{(a,b)}_n(x)^2\\ dx`.
    '''
    if n == 0:
        return (2.**(a+b+1) * gamma(a + 1) * gamma(b + 1)) / gamma(a + b + 2)
    else:
        return (2.**(a+b+1) * gamma(n + a + 1) * gamma(n + b + 1)) / ((2*n + a + b + 1) * gamma(n + a + b + 1) * gamma(n + 1))


def proriolkoornwinderdubiner(d, i, x, out=None):
    '''Evaluation of a Proriol-Koornwinder-Dubiner (PKD) polynomial.

    Args:
        d (int): The spatial dimension.
        i (tuple(int)): Multi-index of length `d`, the degree of the leading monomial of the PKD polynomial in each spatial coordinate.
        x (ndarray): Shape (`N`, `d`), points at which to evaluate the PDK polynomial.
        out (ndarray, optional): Shape (`N`,) array to hold output.

    Returns:
        ndarray: Shape (`N`,), evaluation of the PKD polynomial with leading monomial degrees `i` at the points `x`.  The PKD polynomials
        are orthonormal on the biunit simplex.

    References:
        :cite:`Pror57,KaMc64,Koor75,Dubi91`
    '''
    if (d == 1):
        return jacobi(i[0], x[:, 0], out=out) / jacobinorm2(i[0])**0.5
    else:
        isum = sum(i[0:d-1])
        factor = (1. - x[:, d-1])
        tol = 1.e-10
        nonzero = np.abs(factor) > tol
        if out is None:
            px = np.ndarray(x[:, 0:(d-1)].shape)
        else:
            px = out
        px[nonzero, :] = (x[nonzero, 0:(d-1)] + 1.) * 2. / factor[nonzero, np.newaxis] - 1.
        px[~nonzero, :] = -1.
        pi = proriolkoornwinderdubiner(d-1, i[0:d-1], px)
        pi *= jacobi(i[-1], x[:, d-1], 2*isum + d - 1, 0.)
        pi /= jacobinorm2(i[-1], 2*isum+d-1, 0.)**0.5
        pi *= factor**isum
        pi *= 2.**((d-1)/2)
        return pi


def proriolkoornwinderdubinergrad(d, i, x, out=None, both=False):
    '''Evaluation of the gradient of a Proriol-Koornwinder-Dubiner (PKD) polynomial.

    Args:
        d (int): The spatial dimension.
        i (tuple(int)): Multi-index of length `d`, the degree of the leading monomial of the PKD polynomial in each spatial coordinate.
        x (ndarray): Shape (`N`, `d`), points at which to evaluate the PDK polynomial.
        out (ndarray, optional): Shape (`N`, `d`) array to hold output.
        both: (bool, optional): If ``True``, return both the function value and gradient.

    Returns:
        If ``both=False``, an ndarray with shape (`N`, `d`) containing the
        gradient of the PKD polynomial with leading monomial degrees `i` at the
        points `x`.

        If ``both=True``, a tuple containing two ndarrays, one for the functional value and one for its gradient, in that order.
    '''
    if (d == 1):
        jnorm = jacobinorm2(i[0])**0.5
        if both:
            pkd = jacobi(i[0], x[:, 0]) / jnorm
        out = jacobider(i[0], x[:, 0]).reshape(x.shape)
        out /= jnorm
        if both:
            return (pkd, out)
        return out
    else:
        isum = sum(i[0:d-1])
        factor = (1. - x[:, d-1])
        tol = 1.e-10
        nonzero = np.abs(factor) > tol
        px = np.ndarray(x[:, 0:(d-1)].shape)
        px[nonzero, :] = (x[nonzero, 0:(d-1)] + 1.) * 2. / factor[nonzero, np.newaxis] - 1.
        px[~nonzero, :] = -1.
        pxgradfisum = np.zeros(px.shape + (d,))
        for j in range(d-1):
            pxgradfisum[nonzero, j, j] = 2. * factor[nonzero]**(isum - 1)
            pxgradfisum[nonzero, j, d-1] = (x[nonzero, j] + 1.) * 2. * factor[nonzero]**(isum-2)
            if isum > 0:
                pxgradfisum[~nonzero, j, j] = 2. * factor[~nonzero]**(isum - 1)
                # otherwise, pigrad will be zero
            if isum > 1:
                pxgradfisum[~nonzero, j, d-1] = (x[~nonzero, j] + 1.) \
                                                * 2. * factor[~nonzero]**(isum-2)
                # otherwise, pigrad * pzgrad will be independent of z
        pi, pigrad = proriolkoornwinderdubinergrad(d-1, i[0:d-1], px, both=True)
        pz = jacobi(i[-1], x[:, d-1], 2*isum + d - 1, 0.)
        jnorm = jacobinorm2(i[-1], 2*isum+d-1, 0.)**0.5
        pz /= jnorm
        pzgrad = jacobider(i[-1], x[:, d-1], 2*isum + d - 1, 0.)
        pzgrad /= jnorm
        fisum = factor**isum
        if not (out is None):
            pkdgrad = out
            pkdgrad[:] = 0.
        else:
            pkdgrad = np.zeros(x.shape)
        if both:
            pkd = pi * pz * fisum * 2.**((d-1)/2)
        pkdgrad[:, :] = np.einsum('ij,ijk->ik', pigrad,
                                  pxgradfisum).reshape(x.shape) * pz[:, np.newaxis]
        pkdgrad[:, d-1] += pi * pzgrad * fisum
        if (isum != 0):
            pkdgrad[:, d-1] -= isum * pi * pz * factor**(isum-1)
        pkdgrad *= 2.**((d-1)/2)
        if both:
            return (pkd, pkdgrad)
        return pkdgrad


def proriolkoornwinderdubinerhess(d, i, x):
    '''Evaluation of the gradient of a Proriol-Koornwinder-Dubiner (PKD) polynomial.

    Args:
        d (int): The spatial dimension.
        i (tuple(int)): Multi-index of length `d`, the degree of the leading monomial of the PKD polynomial in each spatial coordinate.
        x (ndarray): Shape (`N`, `d`), points at which to evaluate the PDK polynomial.

    Returns:
        ndarray: Shape (`N`, `d`, `d`) containing the Hessian of the PKD polynomial with leading monomial degrees `i` at the
        points `x`.
    '''
    if (d == 1):
        jnorm = jacobinorm2(i[0])**0.5
        out = jacobider(i[0], x[:, 0], k=2).reshape(x.shape + (1,))
        out /= jnorm
        return out
    else:
        isum = sum(i[0:d-1])
        factor = (1. - x[:, d-1])
        tol = 1.e-10
        nonzero = np.abs(factor) > tol
        px = np.ndarray(x[:, 0:(d-1)].shape)
        px[nonzero, :] = (x[nonzero, 0:(d-1)] + 1.) * 2. / factor[nonzero, np.newaxis] - 1.
        px[~nonzero, :] = -1.
        pxgradf = np.zeros(px.shape + (d,))
        pxgradfisum = np.zeros(px.shape + (d,))
        pxgradfisumm1 = np.zeros(px.shape + (d,))
        pxhessfisum = np.zeros(px.shape + (d, d))
        for j in range(d-1):
            pxgradf[:, j, j] = 2.
            pxgradf[nonzero, j, d-1] = (x[nonzero, j] + 1.) * 2. / factor[nonzero]
            pxgradfisum[nonzero, j, j] = 2. * factor[nonzero]**(isum - 1)
            pxgradfisum[nonzero, j, d-1] = (x[nonzero, j] + 1.) * 2. * factor[nonzero]**(isum-2)
            if isum > 0:
                pxgradfisum[~nonzero, j, j] = 2. * factor[~nonzero]**(isum - 1)
                # otherwise, pigrad will be zero
            if isum > 1:
                pxgradfisum[~nonzero, j, d-1] = (x[~nonzero, j] + 1.) \
                                                * 2. * factor[~nonzero]**(isum-2)
                # otherwise, pigrad * pzgrad will be independent of z
            pxgradfisumm1[nonzero, j, j] = 2. * factor[nonzero]**(isum - 2)
            pxgradfisumm1[nonzero, j, d-1] = (x[nonzero, j] + 1.) * 2. * factor[nonzero]**(isum-3)
            if isum > 1:
                pxgradfisumm1[~nonzero, j, j] = 2. * factor[~nonzero]**(isum - 2)
                # otherwise, pigrad will be zero
            if isum > 2:
                pxgradfisumm1[~nonzero, j, d-1] = (x[~nonzero, j] + 1.) \
                                                  * 2. * factor[~nonzero]**(isum-3)
                # otherwise, pigrad * pzgrad will be independent of z
            pxhessfisum[nonzero, j, d-1, j] = 2. * factor[nonzero]**(isum-2)
            pxhessfisum[nonzero, j, j, d-1] = 2. * factor[nonzero]**(isum-2)
            pxhessfisum[nonzero, j, d-1, d-1] = (x[nonzero, j] + 1.) \
                                                * 4. * factor[nonzero]**(isum-3)
            if isum > 1:
                pxhessfisum[~nonzero, j, d-1, j] = 2. * factor[~nonzero]**(isum-2)
                pxhessfisum[~nonzero, j, j, d-1] = 2. * factor[~nonzero]**(isum-2)
            if isum > 2:
                pxhessfisum[~nonzero, j, d-1, d -
                            1] = (x[~nonzero, j] + 1.) * 4. * factor[~nonzero]**(isum-3)
        pi, pigrad = proriolkoornwinderdubinergrad(d-1, i[0:d-1], px, both=True)
        pihess = proriolkoornwinderdubinerhess(d-1, i[0:d-1], px)
        pz = jacobi(i[-1], x[:, d-1], 2*isum + d - 1, 0.)
        jnorm = jacobinorm2(i[-1], 2*isum+d-1, 0.)**0.5
        pz /= jnorm
        pzgrad = jacobider(i[-1], x[:, d-1], 2*isum + d - 1, 0.)
        pzgrad /= jnorm
        pzhess = jacobider(i[-1], x[:, d-1], 2*isum + d - 1, 0., k=2)
        pzhess /= jnorm
        fisum = factor**isum
        pkdhess = np.zeros(x.shape + (d,))
        pizgradfisum = np.einsum('ij,ijk->ik', pigrad, pxgradfisum).reshape(x.shape)
        pizgradfisumm1 = np.einsum('ij,ijk->ik', pigrad, pxgradfisumm1).reshape(x.shape)

        pkdhess += np.einsum('ijl,ijk,ilm->ikm', pihess, pxgradf,
                             pxgradfisumm1).reshape(x.shape + (d,)) * pz[:, np.newaxis, np.newaxis]
        pkdhess += np.einsum('ij,ijkl->ikl', pigrad, pxhessfisum).reshape(x.shape +
                                                                          (d,)) * pz[..., np.newaxis, np.newaxis]
        pkdhess[:, d-1, d-1] += pi * pzhess * fisum
        if (isum > 1):
            pkdhess[:, d-1, d-1] += (isum-1) * isum * pi * pz * factor**(isum-2)

        temp = pizgradfisum * pzgrad[..., np.newaxis]
        pkdhess[:, :, d-1] += temp
        pkdhess[:, d-1, :] += temp
        if (isum > 0):
            temp = pizgradfisumm1 * (pz * isum)[..., np.newaxis]
            pkdhess[:, :, d-1] -= temp
            pkdhess[:, d-1, :] -= temp
        if (isum > 0):
            pkdhess[:, d-1, d-1] -= 2. * pi * pzgrad * isum * factor**(isum-1)
        pkdhess *= 2.**((d-1)/2)
        return pkdhess


def proriolkoornwinderdubinervandermonde(d, n, x, out=None, C=None):
    '''Evaluation of the Vandermonde matrix of the Proriol-Koornwinder-Dubiner (PKD) polynomials.

    Args:
        d (int): The spatial dimension.
        n (int): The maximum degree of polynomials to include in the Vandermonde matrix.
        x (ndarray): Shape (`N`, `d`), points at which to evaluate the PDK polynomial Vandermonde matrix.
        out (ndarray, optional): Shape (`N`, `\\binom{n+d}{d}`) array to hold output (or, if ``C`` is given, shape
        (`N`, `M`), where `M` is the number of columns of `C`).
        C (ndarray, optional): Shape (`\\binom{n+d}{d}`, `M`) matrix to multiply the PDK Vandermonde matrix by on the right.

    Returns:
        ndarray: Shape (`N`, `\\binom{n+d}{d}`), evaluation of the PKD polynomials with leading monomial degree at most `n` at the points `x`.
        The columns will index the PKD polynomials in lexicographic degree of their leading monomial exponents.

        Or, if `C` is given, shape (`N`, `M`), the product of the Vandermonde matrix with `C`.
    '''
    N = npolys(d, n)
    if out:
        V = out
    else:
        if C is None:
            V = np.ndarray(x.shape[0:-1]+(N,))
        else:
            V = np.ndarray(x.shape[0:-1]+(C.shape[1],))
    if not (C is None):
        V[:] = 0.
    for (k, i) in enumerate(multiindex_up_to(d, n)):
        if C is None:
            V[:, k] = proriolkoornwinderdubiner(d, i, x)
        else:
            V[:, :] += np.einsum('i,j->ij', proriolkoornwinderdubiner(d, i, x), C[k, :])
    return V


def proriolkoornwinderdubinervandermondegrad(d, n, x, out=None, C=None, work=None, both=False):
    '''Evaluation of the gradient of the Vandermonde matrix of the Proriol-Koornwinder-Dubiner (PKD) polynomials.

    Args:
        d (int): The spatial dimension.
        n (int): The maximum degree of polynomials to include in the Vandermonde matrix.
        x (ndarray): Shape (`N`, `d`), points at which to evaluate the PDK polynomial Vandermonde matrix.
        out (ndarray, optional): Shape (`N`, `\\binom{n+d}{d}`, `d`) array to hold output (or, if ``C`` is given, shape
        (`N`, `M`, `d`), where `M` is the number of columns of `C`).
        C (ndarray, optional): Shape (`\\binom{n+d}{d}`, `M`) matrix to multiply the PDK Vandermonde matrix by on the right.

    Returns:
        ndarray: Shape (`N`, `\\binom{n+d}{d}`, `d`), evaluation of the gradient of the PKD Vandermonde matrix at the points `x`.

        Or, if `C` is given, shape (`N`, `M`, `d`), the gradient of the product of the Vandermonde matrix with `C`.
    '''
    N = npolys(d, n)
    if out:
        Vg = out
    else:
        if C is None:
            Vg = np.ndarray(x.shape[0:-1]+(N, d))
        else:
            Vg = np.ndarray(x.shape[0:-1]+(C.shape[1], d))
    if both:
        if C is None:
            V = np.ndarray(x.shape[0:-1]+(N,))
        else:
            V = np.ndarray(x.shape[0:-1]+(C.shape[1],))
    if both:
        V[:] = 0.
    Vg[:] = 0.
    if work is None:
        work = np.ndarray(x.shape[0:-1] + (d,))
    for (k, i) in enumerate(multiindex_up_to(d, n)):
        if both:
            v, work = proriolkoornwinderdubinergrad(d, i, x, out=work, both=True)
        else:
            work = proriolkoornwinderdubinergrad(d, i, x, out=work)
        if C is None:
            if both:
                V[:, k] = v
                Vg[:, k, 0:d] = work
            else:
                Vg[:, k, 0:d] = work
        else:
            if both:
                V[:, :] += np.einsum('i,j->ij', v, C[k, :])
                Vg[:, :, :] += np.einsum('ij,k->ikj', work, C[k, :])
            else:
                Vg[:, :, :] += np.einsum('ij,k->ikj', work, C[k, :])
    if both:
        return (V, Vg)
    return Vg


def proriolkoornwinderdubinervandermondehess(d, n, x, C=None):
    '''Evaluation of the Hessian of the Vandermonde matrix of the Proriol-Koornwinder-Dubiner (PKD) polynomials.

    Args:
        d (int): The spatial dimension.
        n (int): The maximum degree of polynomials to include in the Vandermonde matrix.
        x (ndarray): Shape (`N`, `d`), points at which to evaluate the PDK polynomial Vandermonde matrix.
        C (ndarray, optional): Shape (`\\binom{n+d}{d}`, `M`) matrix to multiply the PDK Vandermonde matrix by on the right.

    Returns:
        ndarray: Shape (`N`, `\\binom{n+d}{d}`, `d`, `d`), evaluation of the
        Hessian of the PKD Vandermonde matrix at the points `x`.

        Or, if `C` is given, shape (`N`, `M`, `d`), the Hessian of the product
        of the Vandermonde matrix with `C`.
    '''
    N = npolys(d, n)
    if C is None:
        Vh = np.ndarray(x.shape[0:-1]+(N, d, d))
    else:
        Vh = np.ndarray(x.shape[0:-1]+(C.shape[1], d, d))
    Vh[:] = 0.
    for (k, i) in enumerate(multiindex_up_to(d, n)):
        work = proriolkoornwinderdubinerhess(d, i, x)
        if C is None:
            Vh[:, k, :, :] = work
        else:
            Vh[:, :, :, :] += np.einsum('ijk,l->iljk', work, C[k, :])
    return Vh




from setuptools import setup, find_packages

python_requires = ">=3.6"
install_requires = ['numpy', 'scipy']
tests_require = ['matplotlib', 'pytest', 'coverage']
docs_require = ['matplotlib', 'sphinx', 'sphinxcontrib-bibtex', 'sphinxcontrib-tikz']
extras_require = {'test': tests_require, 'doc': docs_require}

from itertools import chain
extras_require['all'] = list(chain(*extras_require.values()))

import os
import codecs
import re

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='recursivenodes',
      version=find_version('recursivenodes', '__init__.py'),
      description='Recursively defined interpolation nodes for the simplex',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Toby Isaac',
      author_email='tisaac@cc.gatech.edu',
      packages=['recursivenodes'],
      package_dir={'recursivenodes': 'recursivenodes'},
      python_requires=python_requires,
      install_requires=['numpy', 'scipy'],
      package_data={'recursivenodes':['lebesgue_constants.csv', 'more_lebesgue_constants.csv']},
      tests_require=['matplotlib', 'pytest'],
      license='MIT',
      url="https://tisaac.gitlab.io/recursivenodes/",   # project home page, if any
      project_urls={
          "Bug Tracker": "https://gitlab.com/tisaac/recursivenodes/issues",
          "Source Code": "https://gitlab.com/tisaac/recursivenodes/",
          },
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Intended Audience :: Science/Research",
          "Natural Language :: English",
          "Topic :: Scientific/Engineering :: Mathematics",
      ],
      extras_require=extras_require,
      )




import numpy as np
from scipy.sparse import block_diag
from recursivenodes.polynomials import proriolkoornwinderdubinervandermonde, proriolkoornwinderdubinervandermondegrad, proriolkoornwinderdubinervandermondehess
from recursivenodes.utils import multiindex_up_to, multiindex_equal


def lebesgue(x, d, n, nodes, Vinv=None):
    '''The Lebesgue function in the biunit `d`-simplex with a choice of nodes for degree-`n` polynomials.

    Notes:

        Let `\\{\\boldsymbol{y}_i\\}` be the set of `\\binom{n+d}{d}` nodes.
        The associated shape functions `\\{\\phi_i\\}` are polynomials
        with degree at most `n` such that `\\phi_i(\\boldsymbol{y}_j) = \\delta_{ij}`.

        The Lebesgue function is the sum of the absolute value of the shape functions,
        
        .. math::

            l(\\boldsymbol{x}) = \\sum_{i=1}^{\\binom{n+d}{d}} | \\phi_i(\\boldsymbol{x}) |.

    Args:
    
        x (ndarray): Shape (`N`, `d`), points at which to evaluate the Lebesgue function.
        d (int): The spatial dimension.
        n (int): The polynomial degree that the nodes are unisolvent for.
        nodes (ndarray): Shape (`\\binom{n+d}{d}`, `d`), the nodes defining the shape functions.
        Vinv (ndarray, optional): Shape (`\\binom{n+d}{d}`, `\\binom{n+d}{d}`),
            the inverse of the Vandermonde matrix for the
            Proriol-Koornwinder-Dubiner basis evaluated at the nodes.  These
            polynomials are used as a basis for stable evluation in the biunit
            `d`-simplex, so the inverse of the Vandermonde matrix must be
            computed if it is not already known.

    Returns:
        ndarray: Shape (`N`,), the value of the Lebesgue function for the nodes at the points ``x``.
        

    See Also:
        :func:`recursivenodes.polynomials.proriolkoornwinderdubinervandermonde`.
    '''
    d = nodes.shape[-1]
    if Vinv is None:
        V = proriolkoornwinderdubinervandermonde(d, n, nodes)
        Vinv = np.linalg.inv(V)
    v = proriolkoornwinderdubinervandermonde(d, n, x, C=Vinv)
    return np.abs(v).sum(axis=-1)


def lebesguegrad(x, d, n, nodes, Vinv=None):
    ''' Evaluate the gradient of lebesgue(x, d, n, nodes) w.r.t. x

    x is an (k, d) numpy array for k points in the d-simplex

    nodes is a ((n+d) choose d, d) array of nodes in the d-simplex

    returns a numpy array with the same shape as x

    if the inverse of the Vandermonde matrix of the PKD polynomials evaluated at the nodes is already known,
    it can be passed as Vinv '''
    vec = False
    npoint = -1
    if Vinv is None:
        V = proriolkoornwinderdubinervandermonde(d, n, nodes)
        Vinv = np.linalg.inv(V)
    v, g = proriolkoornwinderdubinervandermondegrad(d, n, x, C=Vinv, both=True)
    s = np.sign(v)
    return (s.reshape(s.shape + (1,)) * g).sum(axis=-2)


def lebesguemin(x, d, n, nodes, Vinv=None):
    '''`-\\sum_j l(\\boldsymbol{x}_j)` and its gradient for the Lebesgue function `l` defined by a set of nodes.

    Args:

        x (ndarray): Shape (`N`, `d`), points at which to evaluate the Lebesgue function.
        d (int): The spatial dimension.
        n (int): The polynomial degree that the nodes are unisolvent for.
        nodes (ndarray): Shape (`\\binom{n+d}{d}`, `d`), the nodes defining the shape functions.
        Vinv (ndarray, optional): See documentation for :func:`lebesgue`.

    Returns:
        (float, ndarray): `-\\sum_j l(\\boldsymbol{x}_j)` and its gradient with respect to ``x``, reshaped to a vector with shape (`Nd`,).
            (This function and return format are intended for use by scipy.optimize.minimize_.)

    .. _scipy.optimize.minimize: https://docs.scipy.org/doc/scipy-1.4.1/reference/generated/scipy.optimize.minimize.html
    '''
    npoint = x.shape[0] // d
    x = x.reshape((npoint, d))
    if Vinv is None:
        V = proriolkoornwinderdubinervandermonde(d, n, nodes)
        Vinv = np.linalg.inv(V)
    v, g = proriolkoornwinderdubinervandermondegrad(d, n, x, C=Vinv, both=True)
    s = np.sign(v)
    return (-np.abs(v).sum(), (-(s.reshape(s.shape + (1,)) * g).sum(axis=-2)).reshape((x.shape[0]*x.shape[1],)))


def lebesgueminhess(x, d, n, nodes, Vinv=None):
    '''The Hessian of `-\\sum_j l(\\boldsymbol{x}_j)`  for the Lebesgue function `l` defined by a set of nodes.

    Args:

        x (ndarray): Shape (`N`, `d`), points at which to evaluate the Lebesgue function.
        d (int): The spatial dimension.
        n (int): The polynomial degree that the nodes are unisolvent for.
        nodes (ndarray): Shape (`\\binom{n+d}{d}`, `d`), the nodes defining the shape functions.
        Vinv (ndarray, optional): See documentation for :func:`lebesgue`.

    Returns:
        scipy.sparse.block_diag_: A block diagonal matrix of shape `(Nd,Nd)`,
            the Hessian of the Lebesgue function with respect to ``x``.
            (This function and return format are intended for use by scipy.optimize.minimize_.)

    .. _scipy.sparse.block_diag: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.block_diag.html
    '''
    npoint = x.shape[0] // d
    x = x.reshape((npoint, d))
    if Vinv is None:
        V = proriolkoornwinderdubinervandermonde(d, n, nodes)
        Vinv = np.linalg.inv(V)
    v = proriolkoornwinderdubinervandermonde(d, n, x, C=Vinv)
    h = proriolkoornwinderdubinervandermondehess(d, n, x, C=Vinv)
    s = np.sign(v)
    lmhess = (-(s.reshape(s.shape + (1, 1)) * h).sum(axis=-3))
    lmhesssp = block_diag(lmhess)
    return lmhesssp


def initial_points(d, n, nodes):
    ''' Given a set of symmetric nodes on the biunit d-simplex to which the equispaced nodes can be deformed without singularities or orientation reversals,
        compute a minimal set of initial guesses which should be sufficient to find all local maxima (up to symmetry) '''
    # enumerate barycentric multiindices that are monotonic: these encompass a subset of the d-simplex whose orbit under vertex permutation covers the whole simplex
    iter_tuples = []
    tuple_to_coord = {}
    for (k, ib) in enumerate(multiindex_equal(d+1, n)):
        tuple_to_coord[ib] = nodes[k, :]
        if all(np.array(ib[0:-1]) <= np.array(ib[1:])):
            iter_tuples.append(ib)
    xinit = []
    # The equispaced nodes induce cells with d types
    # for instance, if d == 1 all cells are just intervals
    #               if d == 2 all cells are either upward or downward triangles
    #               if d == 3 there are upward triangles, octahedra, and downward triangles
    # each type can be characterize as the centroid of a barycentric stencil of width 1, 2, ..., d
    # we start by enumerating these stencils
    for s in range(1, d+1):
        stencil = []
        for i in multiindex_up_to(d, s):
            ib = i + (s-sum(i),)
            # the cell at the center of the stencil is bounded by multiindices containing only 0s and 1s
            if max(ib) == 1:
                # record the stencil as a barycentric stencil, i.e. one whose sum is 0 so that the barycentric condition isn't violated
                ib = i + (-sum(i),)
                stencil.append(ib)
        # print(stencil)
        for t in iter_tuples:  # for every node location (up to symmetry)
            # compute the (equispaced) centroid of the stencil which has this location as its first vertex, in barycentric coordinates
            mp = np.array([((d+1) * t[i] + s) for i in range(d)] + [(d+1)*t[-1] - s*d])
            # if these barycentric coordinates are non-monotonic, then this stencil centroid is in the orbit of another one
            if any(mp[0:-1] > mp[1:]):
                continue
            point = np.zeros(d)
            # compute the centroid for these node coordinates
            for offset in stencil:
                tupl = tuple((t[j] + offset[j]) for j in range(d+1))
                c = tuple_to_coord[tupl]
                point += c
            point /= len(stencil)
            # print(point)
            xinit.append(point)
    xinit = np.array(xinit)
    return (iter_tuples, tuple_to_coord, xinit)


def lebesguemax(d, n, nodes, Vinv=None, solver_options={'gtol': 1.e-3, 'barrier_tol': 1.e-3}, ex_nodes=None, logger=None):
    '''An estimate of the Lebesgue constant `\\Lambda_n^{\\max}(\\Delta^d)`,
    the maximum of the Lebesgue function on the biunit `d`-simplex for a set of
    nodes for polynomials up to degree `n`.

    Notes:

        If `l` is the Lebesgue function defined by the nodes (see :func:`lebesgue`), then

        .. math::
            :label: lmax

            \\Lambda_n^{\\max}(\\Delta^d) = \\max_{\\boldsymbol{x}\\in\\Delta^d} l(\\boldsymbol{x}).

        This constant is significant in quantifying the interpolation quality of a set of nodes because of the error estimate

        .. math::
            
            \\| f - p^n(f) \\|_{\\infty} \\leq (1 + \\Lambda_n^{\\max}) \\inf_{q\\in\\mathcal{P}_n(\\Delta^d)} \\| f - q \\|_{\\infty},

        where `p^n(f)` is the interpolant of `f` defined by the nodes.

        `\\Lambda_n^{\\max}(\\Delta^d)` is nonconvex and only piecewise smooth, with many local maxima.  This
        function is designed to estimate `\\Lambda_n^{\\max}(\\Delta^d)` when
        the nodes have the same lattice-like layout as equispaced nodes (as all of the nodes implemented in this package do, see :mod:`nodes`).

        When this lattice-like structure is present, the maxima always occur in one of the "void"s between nodes in the lattice:

        .. plot::
            :include-source: True

            >>> import numpy as np
            >>> import matplotlib.pyplot as plt
            >>> from recursivenodes import recursive_nodes
            >>> from recursivenodes.nodes import equispaced
            >>> from recursivenodes.utils import coord_map
            >>> from recursivenodes.lebesgue import lebesgue
            >>> x = equispaced(2, 200, domain='biunit')
            >>> nodes = recursive_nodes(2, 6, domain='biunit')
            >>> l = lebesgue(x, 2, 6, nodes)
            >>> x = coord_map(x, 'biunit', 'equilateral')
            >>> ax = plt.gca()
            >>> ax.tricontour(x[:,0], x[:, 1], l, 41)
            <matplotlib.tri.tricontour.TriContourSet object at ...>
            >>> ax.set_aspect('equal')
            >>> plt.show()

        Also, though the Lebesgue function is only piecewise smooth, it is guaranteed to be smooth at the maxima.  An efficient
        way to estimate the `\\Lambda_n^{\\max}(\\Delta^d)` for nodes with lattice-like structure is to try to optimize
        the Lebesgue function from one initial guess at the centroid of each void.  The node sets in this package are also fully symmetric,
        which means that initial guesses can be placed in only a fraction of the voids.  Here, for instance, are the 3D degree 7 equispaced
        nodes and the initial guesses used by ``lebesguemax``:

        .. plot::

            import numpy as np
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import Axes3D
            from recursivenodes import recursive_nodes
            from recursivenodes.nodes import equispaced
            from recursivenodes.lebesgue import initial_points
            from recursivenodes.utils import coord_map
            nodes = equispaced(3, 7, domain='biunit')
            _, _, xinit = initial_points(3, 7, nodes)
            nodes = coord_map(nodes, 'biunit', 'equilateral')
            xinit = coord_map(xinit, 'biunit', 'equilateral')
            ax = plt.gca(projection='3d')
            ax.scatter(nodes[:,0], nodes[:,1], nodes[:,2])
            ax.scatter(xinit[:,0], xinit[:,1], xinit[:,2])
            ax.view_init(None, 225)
            ax.axis('off')
            ax.auto_scale_xyz([-0.7, 0.7], [-0.7, 0.7], [-0.55, 0.85])
            ax.dist = 8
            plt.tight_layout()
            plt.show()


        This function uses scipy.optimize.minimize_(method='trust-constr') to find the local maxima
        from the initial guesses.  While it is not theoretically guaranteed
        that the global maximum is always found in this way, the values of `\\Lambda_n^{\\max}(\\Delta^d)`
        that have been compute with ``lebesguemax`` agree with previously
        computed values in :cite:`BlPo06,LuPo06,Warb06`.

    Args:
    
        d (int): The spatial dimension.
        n (int): The polynomial degree that the nodes are unisolvent for.
        nodes (ndarray): Shape (`\\binom{n+d}{d}`, `d`), the nodes defining the shape functions.
        Vinv (ndarray, optional): See documentation for :func:`lebesgue`.
        solver_options (dict, optional): Options to control the behavior of scipy.optimize.minimize_.
        ex_nodes (ndarray, optional): if ``nodes`` includes nodes that do not touch the boundary of the simplex, ``ex_nodes``
            is a set of nodes that includes extra boundary locations that can be used to help define initial guesses.
        logger (logging.logger, optional): Logger for progress information.

    Returns:
        (float, ndarray): The estimate and estimated argmax of :eq:`lmax`.

    .. _scipy.optimize.minimize: https://docs.scipy.org/doc/scipy-1.4.1/reference/generated/scipy.optimize.minimize.html
    '''

    if Vinv is None:
        V = proriolkoornwinderdubinervandermonde(d, n, nodes)
        Vinv = np.linalg.inv(V)

    if not (ex_nodes is None) and not (ex_nodes is nodes):
        iter_tuples, tuple_to_coord, xinit = initial_points(d, n+d+1, ex_nodes)
    else:
        iter_tuples, tuple_to_coord, xinit = initial_points(d, n, nodes)
    npoints = xinit.shape[0]

    options = solver_options.copy()
    # use the distance between the first two nodes to estimate the initial trust region area
    xscale = np.linalg.norm(nodes[1, :] - nodes[0, :])
    options['initial_tr_radius'] = xscale / 10.

    from scipy.optimize import root, minimize, LinearConstraint
    import scipy.sparse
    constraints = None
    # compute the boundaries of the biunit simplex as bounds for optimization problems
    A = np.concatenate((np.eye(d), np.ones((d,)).reshape((1, d))), axis=0)
    lb = -1 * np.ones((d+1,))
    lb[-1] = -d
    ub = np.ones((d+1,))
    ub[-1] = 2-d
    max_x = xinit.copy()
    # group multiple initial points for better BLAS intensity (although this may cause multiple points to iterate as many iterations as the worst in the group needs)
    block = 8
    for i in range(int(np.ceil(npoints/block))):
        prev = i*block
        last = min(npoints, (i+1)*block)
        size = last - prev
        Ablock = scipy.sparse.kron(scipy.sparse.eye(size), A)
        lbblock = np.tile(lb, size)
        ubblock = np.tile(ub, size)
        constraints = LinearConstraint(Ablock, lbblock, ubblock, keep_feasible=True)
        x = xinit[prev:last, :].reshape((size*d,))
        Ax = Ablock.dot(x)
        assert (Ax >= lbblock).all()
        assert (Ax <= ubblock).all()
        sol = minimize(lebesguemin, x, (d, n, nodes, Vinv), jac=True, hess=lebesgueminhess,
                       method='trust-constr', options=options, constraints=constraints)
        if sol.message != "`gtol` termination condition is satisfied." and sol.message != "`xtol` termination condition is satisfied.":
            import warnings
            warnings.warn(f'scipy.optimize.minimize message: {sol.message}')
            warnings.warn(f'scipy.optimize.minimize function value: {sol.fun}')
            warnings.warn(f'scipy.optimize.minimize solution value: {sol.x}')
        if logger:
            logger.info(f'{last} of {npoints}')
        max_x[prev:last, :] = sol.x.reshape((size, d))

    max_l = lebesgue(max_x, d, n, nodes, Vinv)
    imax = np.argmax(max_l)
    return (max_l[imax], max_x[imax, :])


_lebesguemax_ref = None

def _get_lebesguemax_ref():
    import os
    import csv
    mydir = os.path.dirname(__file__)
    ref = {}
    for filename in ['lebesgue_constants.csv', 'more_lebesgue_constants.csv']:
        datafile = os.path.join(mydir, filename)
        with open(datafile, 'r') as f:
            reader = csv.reader(f)
            for row in reader:
                if len(row) > 0 and row[0][0] == '#': continue
                ref[(row[0], int(row[1]), int(row[2]))] = float(row[3])
    return ref


def lebesguemax_reference(d, n, nodes='recursive'):
    '''Look up previously computed Lebesgue constants (`\\Lambda_n(\\Delta^d)`)
    for node sets implemented in this package and other node sets in the
    literature.

    Args:

        d (int): The spatial dimension.
        n (int): The polynomial degree that the nodes are unisolvent for.
        nodes (str, optional): The name of the node set to query.
            Choices include:

            - ``'recursive'``: :func:`recursivenodes.recursive_nodes`.
            - ``'equispaced'``: :func:`recursivenodes.nodes.equispaced`.
            - ``'warburton'``: :func:`recursivenodes.nodes.warburton`.
            - ``'blyth_luo_pozrikidis'``: :func:`recursivenodes.nodes.blyth_luo_pozrikidis`.
            - ``'Chen-Babuska'``: Chen-Babuška type nodes :cite:`ChBa95,ChBa96`, computed in :cite:`Roth05`.
            - ``'Fekete'``: Fekete nodes, computed in :cite:`Roth05`.
            - ``'Lebesgue'``: Lebesgue-constant-minimizing nodes, computed in :cite:`Roth05`.
            - ``'Heinrichs-asymmetric'``: Lebesgue-constant-minimizing nodes :cite:`Hein05`.
            - ``'Heinrichs-symmetric'``: Symmetric Lebesgue-constant-minimizing nodes :cite:`Hein05`.
            - ``'Heshaven'``: Hesthaven electrostatic nodes :cite:`Hest98`.
            - ``'Heshaven-Teng'``: Hesthaven-Teng electrostatic nodes :cite:`Hest98`.
            - ``'Rapetti-Sommariva-Vianello-leb'``: Lebesgue-constant-minimizing nodes, computed in :cite:`RaSV12`.
            - ``'Rapetti-Sommariva-Vianello-lebgl'``: Lebesgue-constant-minimizing nodes wih Lobatto-Gauss-Legendre edges, computed in :cite:`RaSV12`.
            - ``'Rapetti-Sommariva-Vianello-lebgls'``: Symmetric Lebesgue-constant-minimizing nodes wih Lobatto-Gauss-Legendre edges, computed in :cite:`RaSV12`.

    Returns:
        float: `\\Lambda_n(\\Delta^d)` for that node set.

    Raises:
        KeyError: The node set is unknown, or `\\Lambda_n(\\Delta^d)` has not been precomputed for it.
    '''
    global _lebesguemax_ref
    if _lebesguemax_ref is None:
        _lebesguemax_ref = _get_lebesguemax_ref()
    return _lebesguemax_ref[(nodes, d, n)]


def add_lebesguemax_to_parser(parser):
    from recursivenodes.nodes import add_nodes_to_parser
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='show verbose computation progress')
    parser.add_argument('-t', '--tol', type=float, nargs='?', default=1.e-3,
                        help='convergence tolerance for gradient convergence')
    add_nodes_to_parser(parser)


def lebesguemax_from_args(args):
    from recursivenodes.nodes import nodes_from_args
    from recursivenodes.utils import coord_map
    d, n, domain, family, nodes, ex_nodes = nodes_from_args(args, expanded=True)
    if domain != 'biunit':
        new_nodes = coord_map(nodes, domain, 'biunit')
        if ex_nodes is nodes:
            new_ex_nodes = new_nodes
        else:
            new_ex_nodes = coord_map(ex_nodes, domain, 'biunit')
        nodes = new_nodes
        ex_nodes = new_ex_nodes
    solver_options={'gtol': args.tol, 'barrier_tol': args.tol}
    logger = None
    if args.verbose:
        import logging
        logging.basicConfig()
        logger = logging.getLogger('lebesguemax')
        logger.setLevel('INFO')
    lmaxreg = lebesguemax(d, n, nodes, ex_nodes=ex_nodes, solver_options=solver_options,
                          logger=logger)
    return lmaxreg


if __name__ == '__main__':
    from argparse import ArgumentParser

    parser = ArgumentParser(
        description='Compute the maximum of the Lebesgue function on the d-simplex for a node set implemented in nodes.py')
    add_lebesguemax_to_parser(parser)
    args = parser.parse_args()
    (max_l, max_x) = lebesguemax_from_args(args)
    if args.verbose:
        print((max_l, max_x))
    else:
        print(max_l)
